<template name="file-header">
    /*
        Generated by sbus code generator

        Copyright (C) 2017 Red Hat

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 3 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program.  If not, see <http://www.gnu.org/licenses/>.
    */

    #include "${sbus-path}/sbus_interface_declarations.h"
    #include "${header:symbols}"

</template>

<template name="annotations">
    const struct sbus_annotation
    _sbus_annotations_${token}[] = {
        <loop name="annotation">
        {.name = "${annotation-name}", .value = ${annotation-value}},
        </loop>
        {NULL}
    };

</template>

<template name="method">
    const struct sbus_method_arguments
    _sbus_args_${token} = {
        .input = (const struct sbus_argument[]){
            <loop name="input">
            {.type = "${arg-type}", .name = "${arg-name}"},
            </loop>
            {NULL}
        },
        .output = (const struct sbus_argument[]){
            <loop name="output">
            {.type = "${arg-type}", .name = "${arg-name}"},
            </loop>
            {NULL}
        }
    };

</template>

<template name="signal">
    const struct sbus_argument
    _sbus_args_${token}[] = {
        <loop name="input">
        {.type = "${arg-type}", .name = "${arg-name}"},
        </loop>
        {NULL}
    };

</template>
