#define SSSD_TEST_CERT_SSH_KEY_0008 "AAAAB3NzaC1yc2EAAAADAQABAAABAQCnbAWrNFBNKUDLZ/2dh69WYYFlEZ3p/bl2FO9YdMwNC3ewLmBv3/LQK013bBgv8nQ3R/Ko83pQrAXecN1eTFcATIIVIUii20mlKCOVpJDT3x4XLne6tVImualjftBQC2O6XoQAl3utKszw2CygjnixtFUBPV/0Ycmc+1zCzhQax8cGSCFUmyP0ZydIpERENUFQiKPDk/b0ztxI1sEuktu3lZkOjnRAoRFBQOx+K4jzZezGUULz6HPXtyKLOVLa/PaPfLDrQe/s5DITOIYIXOfCM9iFdQbYoAMoTT+H2vNcFuqIZZV/024HlUGhLqAMH1WiCiaOL33JCABU1b1hokEV"
