# Constants
ds_instance_name = 'example1'
ds_instance_name1 = 'example2'
ds_instance_name2 = 'example3'
ds_suffix = 'dc=example,dc=test'
krb_realm = 'EXAMPLE.TEST'
ds_rootdn = 'cn=Directory Manager'
ds_rootpw = 'Secret123'
