#define SSSD_TEST_CERT_SSH_KEY_0003 "AAAAB3NzaC1yc2EAAAADAQABAAABAQDb6pvC57nKEY/yk6PXmcbIHz3lmUXqL1OPdJrBXFgs57AxIPvbXqWNJi4/c82+nsdmzKunrU1B+ZkUpoNd9eT/c3y7RaZ7UeGBqUhUBRXd/zllQmGnDfwPLIirhK7+UTjOnez6WSPHzKWS36ZZG6aZXj//lc9X/8xOb5kr/Upn4AFTHkhirOpzMJY2T9pEDj/gx8o+FUgHnw5oWlarnPCod3xpn2RtNsugpTLdtDMhhPabaySUJdNeHYRfP5UUqjEr9yvWDuwfjW2y5TMoqZreEg6RSmya6BDm9nuwqk07NAK581AvFpSNvy1H8PL1BckXe6oxskPUM2rDXscGCuOn"
