<template name="file-header">
    /*
        Generated by sbus code generator

        Copyright (C) 2017 Red Hat

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 3 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program.  If not, see <http://www.gnu.org/licenses/>.
    */

    #ifndef ${file-guard}
    #define ${file-guard}

    #include <stdint.h>
    #include <stdbool.h>

    <loop name="custom-type-header">
    #include "${custom-type-header}"
    </loop>

</template>

<template name="properties">
    struct sbus_all_${token} {
        <loop name="property">
        struct {
            bool is_set;
            ${input-type} value;
        } ${name};
        </loop>
    };

</template>

<template name="file-footer">
    #endif /* ${file-guard} */
</template>
