/*
    Authors:
        Pavel Březina <pbrezina@redhat.com>

    Copyright (C) 2013 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef COMMON_MOCK_SYSDB_OBJECTS_H_
#define COMMON_MOCK_SYSDB_OBJECTS_H_

#include <talloc.h>

#include "util/util.h"
#include "providers/ldap/sdap.h"

struct sysdb_attrs *
_mock_sysdb_object(TALLOC_CTX *mem_ctx,
                   const char *base_dn,
                   const char *name,
                   ...);

#define mock_sysdb_object(mem_ctx, base_dn, name, ...) \
    _mock_sysdb_object(mem_ctx, base_dn, name, ##__VA_ARGS__, NULL)

struct sysdb_attrs *
mock_sysdb_group_rfc2307bis(TALLOC_CTX *mem_ctx,
                            const char *base_dn,
                            gid_t gid,
                            const char *name,
                            const char **members);

struct sysdb_attrs *
mock_sysdb_user(TALLOC_CTX *mem_ctx,
                const char *base_dn,
                uid_t uid,
                const char *name);

#endif /* COMMON_MOCK_SYSDB_OBJECTS_H_ */
