#define SSSD_TEST_ECC_CERT_SSH_KEY_0001 "AAAAE2VjZHNhLXNoYTItbmlzdHAzODQAAAAIbmlzdHAzODQAAABhBCF5kUGe3NwBEDnYZtrtzTUESUVF2zUXjf0qi0U6gPHh8l/ruGYPr1sq5NtYu9zuXct/ZMp0jkYKn+71I+Ea297QJpib0dyjHCnbJieGrey2hb59qv//iHeZDFT+Eb2mPQ=="
