#
# Miscellaneous routines.
#
# Copyright (C) 2014 Red Hat
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

if [ -z ${_MISC_SH+set} ]; then
declare -r _MISC_SH=

# Remove files and directories recursively, forcing write permissions on
# directories.
# Args: path...
function rm_rf_ro()
{
    chmod -Rf u+w -- "$@" || true
    rm -Rf -- "$@"
}

fi # _MISC_SH
